unit Formpage;

{ $DEFINE DontUseThread}

{
This helps to save resources using a notebook.
Use this software as freeware. The only thing I ask you is that
if you find any bugs, please E-Mail me!
Christoph Kirchner, ckirchner@geocities.com

Based on a program by David Obber:
> Thanks to 3k Computer Consultancy for useful hints.
> David Obber, MC8783@mclink.it

Changed by Neil Peters, petersn@cadvision.com:
  Bugfixes and support for the PageControl object


------------------------------------------------------------------
							DISCLAIMER
------------------------------------------------------------------
Use this software at your own risk. I won't be held responsible for
anything that goes wrong.

}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, TabNotBk, ComCtrls;

const
  MAXNBKPAGES = 16;  { Maximum number of pages = 16 }
  UM_SIZE = WM_USER + 1100;        // may have problem here, related to non-unique message number conflicts or something -
                                   // sorry I can not figure the problem out yet.
  UM_SIZE_MAXIMIZED = WM_USER + 1101;
  UM_SIZE_RESTORED = WM_USER + 1102;
  UM_SETFOCUS = WM_USER + 1103;
  UM_RESETCURSOR = WM_USER + 1104;
  UM_DestroyThread = WM_USER + 1105;

const
  StartThreadPriority = tpLower; { or  tpIdle ? }


type

  TScaledBy = record
    M, D: Integer;
  end;

  TScaleForm = class(TForm)
  private
    fmh, fmw, fml, fmt : word;
    nOldWidth, nOldHeight: integer;
    AllowScale: WordBool;
    AllowResize: WordBool;
    procedure mymax(var m: TWMGETMINMAXINFO); message wm_getminmaxinfo;
    procedure AfterSIZE_MAXIMIZED(var message: TMessage); message UM_SIZE_MAXIMIZED;
    procedure AfterSIZE_RESTORED(var message: TMessage); message UM_SIZE_RESTORED;
    procedure ScaleComponents(Reason, M, D: Integer);
  protected
    procedure WMSize(var message: TWMSize); message WM_SIZE;
    procedure ChangeScale(M, D: Integer); override;
  public
    ScaledBy: TScaledBy;
    constructor Create(AOwner: TComponent); override;
  published
    property maxheight : word read fmh write fmh;
    property maxwidth  : word read fmw write fmw;
    property maxleft   : word read fml write fml;
    property maxtop    : word read fmt write fmt;
  end;

  TThreadForm = class;

  TPseudoThread = class
  private
    FHandle: THandle;
    FThreadID: THandle;
    FTerminated: Boolean;
    FSuspended: Boolean;
    FFreeOnTerminate: Boolean;
    FPriority: TThreadPriority;
    FReturnValue: Integer;
  protected
    Destroying: Boolean;
    procedure Synchronize(Method: TThreadMethod);
    property ReturnValue: Integer read FReturnValue write FReturnValue;
    property Terminated: Boolean read FTerminated;
  public
    constructor Create(CreateSuspended: Boolean);
{   destructor Destroy; override; }
    procedure Resume;
    procedure Suspend;
    procedure Terminate;
    procedure Execute; virtual; abstract;
    property FreeOnTerminate: Boolean read FFreeOnTerminate write FFreeOnTerminate;
    property Handle: THandle read FHandle;
    property Priority: TThreadPriority read FPriority write FPriority;
    property Suspended: Boolean read FSuspended;
    property ThreadID: THandle read FThreadID;
  end;


{$IFDEF DontUseThread}
  TStartThread = class(TPseudoThread)
{$ELSE DEF DontUseThread}
  TStartThread = class(TThread)
{$ENDIF DEF DontUseThread}
  private
    procedure SetThreadFormVisible(Value: Boolean);
  protected
    FThreadForm: TThreadForm;
    FThreadFormVisible: Boolean;
    PleaseDoNotShowThreadForm: Boolean;
  public
    constructor Create(AThreadForm: TThreadForm);
    destructor Destroy; override;
    property ThreadFormVisible: Boolean
      read FThreadFormVisible write SetThreadFormVisible;
  end;

{$IFDEF DontUseThread}
  TShowThread = class(TPseudoThread)
{$ELSE DEF DontUseThread}
  TShowThread = class(TThread)
{$ENDIF DEF DontUseThread}
  protected
    FThreadForm: TThreadForm;
    function WaitForStartThread: Boolean; { false: terminated }
  public
    constructor Create(AThreadForm: TThreadForm);
    destructor Destroy; override;
  end;

 TDefaultShowThread = class(TShowThread)
  protected
    procedure Execute; override;
  end;

  TThreadForm = class(TScaleForm)
  private
    FStartThread: TStartThread;
    FShowThread: TShowThread;
    FShowGrayed: Boolean;
    procedure OnUM_RESETCURSOR(var message: TMessage); message UM_RESETCURSOR;
    procedure OnUM_DestroyThread(var message: TMessage); message UM_DestroyThread;
  protected
    procedure DoHide; override;
    procedure DoShow; override;
    procedure Paint; override;
    procedure SetStartThread(Value: TStartThread); virtual;
    procedure SetShowThread(Value: TShowThread); virtual;
    procedure SetShowGrayed(Value: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property StartThread: TStartThread read FStartThread write SetStartThread;
    property ShowThread: TShowThread read FShowThread write SetShowThread;
    property ShowGrayed: Boolean write SetShowGrayed;
  end;

  TPageForm = class(TThreadForm)
  private
    procedure OnUM_SETFOCUS(var message: TMessage); message UM_SETFOCUS;
  protected
    WinCtrl: TWinControl;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    ScaledBy: TScaledBy;
    constructor Create(AOwner: TComponent); override;
  end;


  TPageFormClass = class of TPageForm;

  TPageFormInfo = class(TComponent)
  private
    FFormContainer: TComponent;
    FormCreated: TPageForm;
    FormDestroy: TNotifyEvent;
    FDontDestroy: Boolean;
    procedure OnFormDestroy(Sender: TObject);
  protected
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
  public
    FormType: TPageFormClass;
    function  CreateForm: TPageForm;
    procedure DestroyForm;
    procedure ShowForm;
    procedure HideForm;
  published
    property FormContainer: TComponent
      read FFormContainer write FFormContainer;
    property DontDestroy: Boolean read FDontDestroy write FDontDestroy;
  end;

  TFormPages = class(TComponent)
  private
    FormInfo: array [0..MAXNBKPAGES-1] of TPageFormInfo;
  protected
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure HideForm(Page: Integer);
    procedure DestroyForm(Page: Integer);
    procedure ChangeForm(Handle: HWND; OldPage, NewPage: Integer);
  public
    Constructor Create(AOwner:Tcomponent); override;
    Destructor Destroy; override;
    procedure AddPage(Page: Integer; AFormContainer: TComponent;
      AFormType: TPageFormClass);
    function  CreateForm(Page: Integer): TPageForm;
    procedure ShowForm(Page: Integer);
  end;

  TTabbedNotebookExternalPages = class(TFormPages)
  private
    FTabbedNotebook: TTabbedNotebook;
    FTabbedNotebookChange: TPageChangeEvent;
    ParentOnShow: TNotifyEvent;
    Constructor Create(AOwner:Tcomponent); override;
    function  GetTabbedNotebook: TTabbedNotebook;
    Procedure SetTabbedNotebook(value : TTabbedNotebook);
    procedure ReleaseTabbedNotebook;
    procedure TabbedNotebookChange(Sender: TObject; NewTab: Integer;
      var AllowChange: Boolean);
    procedure OnParentShow(Sender: TObject);
  protected
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
  public
    procedure CreateCurrentPage;
  published
    property TabbedNotebook: TTabbedNotebook read GetTabbedNotebook write SetTabbedNotebook;
    property PageFormInfo00: TPageFormInfo
      read  FormInfo[0] write FormInfo[0];
    property PageFormInfo01: TPageFormInfo
      read  FormInfo[1] write FormInfo[1];
    property PageFormInfo02: TPageFormInfo
      read  FormInfo[2] write FormInfo[2];
    property PageFormInfo03: TPageFormInfo
      read  FormInfo[3] write FormInfo[3];
    property PageFormInfo04: TPageFormInfo
      read  FormInfo[4] write FormInfo[4];
    property PageFormInfo05: TPageFormInfo
      read  FormInfo[5] write FormInfo[5];
    property PageFormInfo06: TPageFormInfo
      read  FormInfo[6] write FormInfo[6];
    property PageFormInfo07: TPageFormInfo
      read  FormInfo[7] write FormInfo[7];
    property PageFormInfo08: TPageFormInfo
      read  FormInfo[8] write FormInfo[8];
    property PageFormInfo09: TPageFormInfo
      read  FormInfo[9] write FormInfo[9];
    property PageFormInfo10: TPageFormInfo
      read  FormInfo[10] write FormInfo[10];
    property PageFormInfo11: TPageFormInfo
      read  FormInfo[11] write FormInfo[11];
    property PageFormInfo12: TPageFormInfo
      read  FormInfo[12] write FormInfo[12];
    property PageFormInfo13: TPageFormInfo
      read  FormInfo[13] write FormInfo[13];
    property PageFormInfo14: TPageFormInfo
      read  FormInfo[14] write FormInfo[14];
    property PageFormInfo15: TPageFormInfo
      read  FormInfo[15] write FormInfo[15];
  end;

  TPageControlExternalPages = class(TFormPages)
  private
    FPageControl : TPageControl ;
    FPageControlChange: TNotifyEvent ;
    FPageControlChanging : TTabChangingEvent  ;   //
    ParentOnShow: TNotifyEvent;
    FPreviousPage: TTabSheet;  // added property
    Constructor Create(AOwner:Tcomponent); override;
    function  GeTPageControl : TPageControl ;
    Procedure SeTPageControl (value : TPageControl );
    procedure ReleasePageControl ;
    procedure PageControlChange(Sender: TObject) ;
    procedure PageControlChanging(Sender: TObject;
      var AllowChange: Boolean);   //
    procedure OnParentShow(Sender: TObject);
  protected
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
  public
    procedure CreateCurrentPage;
    property PreviousPage: TTabSheet read FPreviousPage write FPreviousPage;
  published
    property PageControl : TPageControl  read GeTPageControl  write SeTPageControl ;
    property PageFormInfo00: TPageFormInfo
      read  FormInfo[0] write FormInfo[0];
    property PageFormInfo01: TPageFormInfo
      read  FormInfo[1] write FormInfo[1];
    property PageFormInfo02: TPageFormInfo
      read  FormInfo[2] write FormInfo[2];
    property PageFormInfo03: TPageFormInfo
      read  FormInfo[3] write FormInfo[3];
    property PageFormInfo04: TPageFormInfo
      read  FormInfo[4] write FormInfo[4];
    property PageFormInfo05: TPageFormInfo
      read  FormInfo[5] write FormInfo[5];
    property PageFormInfo06: TPageFormInfo
      read  FormInfo[6] write FormInfo[6];
    property PageFormInfo07: TPageFormInfo
      read  FormInfo[7] write FormInfo[7];
    property PageFormInfo08: TPageFormInfo
      read  FormInfo[8] write FormInfo[8];
    property PageFormInfo09: TPageFormInfo
      read  FormInfo[9] write FormInfo[9];
    property PageFormInfo10: TPageFormInfo
      read  FormInfo[10] write FormInfo[10];
    property PageFormInfo11: TPageFormInfo
      read  FormInfo[11] write FormInfo[11];
    property PageFormInfo12: TPageFormInfo
      read  FormInfo[12] write FormInfo[12];
    property PageFormInfo13: TPageFormInfo
      read  FormInfo[13] write FormInfo[13];
    property PageFormInfo14: TPageFormInfo
      read  FormInfo[14] write FormInfo[14];
    property PageFormInfo15: TPageFormInfo
      read  FormInfo[15] write FormInfo[15];
  end;

procedure Register;

implementation



procedure TScaleForm.ScaleComponents(Reason, M, D: Integer);
var
  i: Integer;
begin
{  for i := ControlCount-1 downto 0 do }
  for i := 0 to ControlCount-1 do
  begin
    if (Controls[i] is TWinControl) then
      if (Controls[i] is TScaleForm) then
      begin
        with Controls[i] as TScaleForm do
        begin
          ScaleBy(M, D); {}
          if (Reason = SIZE_RESTORED) then
          begin
            ScaledBy.M := 1;
            ScaledBy.D := 1;
          end
          else
          begin
            ScaledBy.M := M;
            ScaledBy.D := D;
          end;
        end;
      end
      else
        with Controls[i] as TWinControl do ScaleBy(M, D);
  end;
end;

procedure TScaleForm.AfterSIZE_MAXIMIZED(var message: TMessage);
begin
      if (Width <> nOldWidth) then
        ScaleComponents(SIZE_MAXIMIZED, Width, nOldWidth);
      ScaledBy.M := Width;
      ScaledBy.D := nOldWidth;
      nOldWidth := Width;
end;

procedure TScaleForm.AfterSIZE_RESTORED(var message: TMessage);
begin
      if (ScaledBy.M <> 1) or (ScaledBy.D <> 1) then
        ScaleComponents(SIZE_RESTORED, ScaledBy.D, ScaledBy.M);
      ScaledBy.M := 1;
      ScaledBy.D := 1;
      nOldWidth := Width;
end;

procedure TScaleForm.WMSize(var message: TWMSize);
begin
  if (not (csDesigning in ComponentState)) and (AllowScale) then
  begin
    if not AllowResize then
    begin
      message.Result := 0;
      exit;
    end;
    if (WindowState = wsMinimized) then
      exit;
    if (message.SizeType = SIZE_RESTORED) then
      PostMessage(Handle, UM_SIZE_RESTORED, 0, 0);
    if (message.SizeType = SIZE_MAXIMIZED) then
      PostMessage(Handle, UM_SIZE_MAXIMIZED, 0, 0);
    PostMessage(Handle, UM_SIZE, message.SizeType, 0);
(*
    if (message.SizeType = SIZE_MAXIMIZED) then
    begin { Window has been maximized: }
      if (Width <> nOldWidth) then
        ScaleComponents(Width, nOldWidth);
      ScaledBy.M := Width;
      ScaledBy.D := nOldWidth;
      exit;
    end;
    if (message.SizeType = SIZE_RESTORED) then
    begin { Window has been resized: }
      if (ScaledBy.M <> 1) or (ScaledBy.D <> 1) then
        ScaleComponents(ScaledBy.D * 2, ScaledBy.M * 2);
      ScaledBy.M := 1;
      ScaledBy.D := 1;
      exit;
    end;
*)
  end
  else
  begin
    nOldWidth := Width;
    nOldHeight := Height;
  end;
end;


procedure TScaleForm.mymax(var m : TWMGETMINMAXINFO);
begin
  m.minmaxinfo^.ptmaxsize.x := fmw;
  m.minmaxinfo^.ptmaxsize.y := fmh;
  m.minmaxinfo^.ptmaxposition.x := fml;
  m.minmaxinfo^.ptmaxposition.y := fmt;
end;

constructor TScaleForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fmw := screen.width;
  fmh := screen.height;
  fmt := 0;
  fml := 0;
  ScaledBy.M := 1;
  ScaledBy.D := 1;
  nOldWidth := Width;
  nOldHeight := Height;
  if (AOwner is TScaleForm) then
  begin
    with TScaleForm(AOwner).ScaledBy do begin
      if (M <> 1) and (D <> 1) and (M <> 0) and (D <> 0) then
      begin { the owner is already scaled: }
        nOldWidth := Longint(nOldWidth) * D div M;
        nOldHeight := Longint(nOldHeight) * D div M;
        AllowScale := true;
    AllowScale := false; { the owner will do the scaling for us }
{        ScaleBy(M, D); {}
        ScaledBy.M := M;
        ScaledBy.D := D;
      end;
    end;
    AllowScale := false; { the owner will do the scaling for us }
  end
  else
    AllowScale := true;
  AllowResize := true;
end;

procedure TScaleForm.ChangeScale(M, D: Integer);
{
var
  ScaledByM: Integer; }
begin
  inherited ChangeScale(M, D);
{
  ScaledByM := ScaledByM;
  ScaledBy.M := M div ScaledBy.D;
  ScaledBy.D := D div ScaledByM;
}
end;



{ TStartThread ----------------------------------------------------------------}

constructor TStartThread.Create(AThreadForm: TThreadForm);
begin
  inherited Create(true);
  FreeOnTerminate := True;
  Priority := StartThreadPriority;
  FThreadForm := AThreadForm;
  FThreadFormVisible := false;
  PleaseDoNotShowThreadForm := false;
  Resume;
{$IFDEF DontUseThread}
  PostMessage(FThreadForm.Handle, UM_DestroyThread, 0, Longint(self));
{$ENDIF DEF DontUseThread}
end;

destructor TStartThread.Destroy;
begin
  Terminate;
  Priority := tpHigher; { or tpHighest ? }
  try
    if Assigned(FThreadForm) then
    begin
      if Assigned(FThreadForm.StartThread) and (FThreadForm.StartThread = self) then
        FThreadForm.StartThread := nil;
      if Assigned(FThreadForm.FShowThread) then
        FThreadForm.FShowThread.Resume;
    end;
  except end;
  inherited Destroy;
end;

procedure TStartThread.SetThreadFormVisible(Value: Boolean);
begin  { called by Main-Thread }
  if Terminated or (FThreadFormVisible = Value) then
    exit;
  if Value then
    while PleaseDoNotShowThreadForm and (not Terminated) do
    begin
(*
      if (GetCurrentThreadID = MainThreadID) and
         Assigned(ApplicationProcessMessages) then
      begin
        while PleaseDoNotShowThreadForm and (not Terminated) do
        begin
          ApplicationProcessMessages; { time for main-thread }
          if PleaseDoNotShowThreadForm and (not Terminated) then
            Sleep(50); { time for other threads }
        end;
      end
      else
      begin
*)
        while PleaseDoNotShowThreadForm and (not Terminated) do
        begin
          Sleep(50);
        end;
//    end;
    end;
  FThreadFormVisible := Value;
end;

{ TShowThread ----------------------------------------------------------------}

constructor TShowThread.Create(AThreadForm: TThreadForm);
begin
  inherited Create(true);
  FreeOnTerminate := true; {}
  FThreadForm := AThreadForm;
try
  Screen.Cursor := crHourglass;    { Sanduhr-Cursor anzeigen }
{
  if Assigned(FThreadForm) then
    PostMessage(FThreadForm.Handle, UM_RESETCURSOR, 0, crHourglass);
}
  if Assigned(FThreadForm) and Assigned(FThreadForm.FStartThread) then
  begin
    FThreadForm.FStartThread.Priority := tpHigher; { or tpHighest or tpTimeCritical ? }
    if Assigned(FThreadForm.FStartThread) then
      FThreadForm.FStartThread.Resume;
  end
  else
    Resume;
except
  Resume;
end;
{$IFDEF DontUseThread}
  PostMessage(FThreadForm.Handle, UM_DestroyThread, 0, Longint(self));
{$ENDIF DEF DontUseThread}
{  Priority := tpHigher; {}
end;

destructor TShowThread.Destroy;
begin
{  Screen.Cursor := crDefault; {}
{ SetCursor(LoadCursor(0, IDC_ARROW)); {}
  Terminate;
  if Assigned(FThreadForm) then
    PostMessage(FThreadForm.Handle, UM_RESETCURSOR, 0, crDefault);
  Priority := tpHigher; { or tpHighest ? }
  try
    if Assigned(FThreadForm) and Assigned(FThreadForm.FShowThread) and
       (FThreadForm.FShowThread = self) then
    begin
      FThreadForm.ShowThread := nil;
    end;
  except end;
  inherited Destroy;
end;

function TShowThread.WaitForStartThread: Boolean; { false: terminated }
{var
  OldCursor: HCursor;}
begin
{ OldCursor := SetCursor(LoadCursor(0, IDC_WAIT)); {}
  Screen.Cursor := crHourglass;    { Sanduhr-Cursor anzeigen }
  while Assigned(FThreadForm) and Assigned(FThreadForm.FStartThread) and
        (not Terminated) do
  begin
    Sleep(50); { time for other threads }
  end;
{ SetCursor(OldCursor); {}
{  Screen.Cursor := crDefault; {}
  Result := not Terminated;
end;


{ TDefaultShowThread ---------------------------------------------------------}

procedure TDefaultShowThread.Execute;
begin
  if not WaitForStartThread then exit;
end;


{ TThreadForm -----------------------------------------------------------------}

constructor TThreadForm.Create(AOwner: TComponent);
begin
  FStartThread := nil;
  FShowThread := nil;
  FShowGrayed := false;
  inherited Create(AOwner);
end;

destructor TThreadForm.Destroy;
begin
  try
    if Assigned(FStartThread) then
    begin
      FStartThread.FThreadForm := nil;
//    if not FStartThread.Destroying then
        FStartThread.Free;
    end;
  finally
    FStartThread := nil;
  end;
  try
    if Assigned(FShowThread) then
    begin
      FShowThread.FThreadForm := nil;
//    if not FShowThread.Destroying then
        FShowThread.Free;
    end;
  finally
    FShowThread := nil;
  end;
  inherited Destroy;
end;

procedure TThreadForm.DoHide;
begin
  if (FStartThread <> nil) then
  begin
    if not FStartThread.Terminated then
    begin
      if (FStartThread <> nil) then
        FStartThread.Priority := StartThreadPriority;
      if (FStartThread <> nil) then
        FStartThread.ThreadFormVisible := false;
    end;
  end;
  inherited;
end;

procedure TThreadForm.DoShow;
begin
  try
  { if FShowThread is running then terminate it,
    because we will create a new one: }
    if Assigned(FShowThread) then
    begin
      FShowThread.FThreadForm := nil;
      FShowThread.Terminate;
    end;
  finally
    FShowThread := nil;
  end;
  { if FStartThread is running then raise priority,
    because FShowThread will wait until it ends: }
  if (FStartThread <> nil) then
  begin
    if not FStartThread.Terminated then
    begin
      if (FStartThread <> nil) then
        FStartThread.Priority := tpHighest;	{ or tpTimeCritical ? }
      if (FStartThread <> nil) then
        FStartThread.Resume;
      if (FStartThread <> nil) then
        FStartThread.ThreadFormVisible := true;
    end;
  end;
  inherited;
{ if no FShowThread is created, create default: }
  if not Assigned(FShowThread) then
    FShowThread := TDefaultShowThread.Create(self);
end;

procedure TThreadForm.OnUM_RESETCURSOR(var message: TMessage);
begin
  Screen.Cursor := message.LParam;
end;

procedure TThreadForm.OnUM_DestroyThread(var message: TMessage);
begin
  try
    if Assigned(FShowThread) and (Longint(FShowThread) = Message.lParam) then
    begin
      FShowThread.Terminate;
    end;
  finally
    FShowThread := nil;
  end;
  try
    if Assigned(FStartThread) and (Longint(FStartThread) = Message.lParam) then
    begin
      FStartThread.Terminate;
    end;
  finally
    FStartThread := nil;
  end;
end;

procedure TThreadForm.SetStartThread(Value: TStartThread);
begin
  if (FStartThread <> Value) then
  begin
    if (Value <> nil) then
      ShowGrayed := true
    else
      if (FShowThread = nil) then
        ShowGrayed := false;
    FStartThread := Value;
  end;
end;

procedure TThreadForm.SetShowThread(Value: TShowThread);
begin
  if (FShowThread <> Value) then
  begin
    if (Value = nil) then
      ShowGrayed := false;
    FShowThread := Value;
  end;
end;

procedure TThreadForm.SetShowGrayed(Value: Boolean);
begin
  if (FShowGrayed <> Value) then
  begin
    Invalidate;
    FShowGrayed := Value;
  end;
end;

procedure ShadeRect(DC: HDC; const Rect: TRect);
const
  HatchBits: array[0..7] of Word = ($11, $22, $44, $88, $11, $22, $44, $88);
var
  Bitmap: HBitmap;
  SaveBrush: HBrush;
  SaveTextColor, SaveBkColor: TColorRef;
begin
  Bitmap := CreateBitmap(8, 8, 1, 1, @HatchBits);
  SaveBrush := SelectObject(DC, CreatePatternBrush(Bitmap));
  SaveTextColor := SetTextColor(DC, clWhite);
  SaveBkColor := SetBkColor(DC, clBlack);
  with Rect do PatBlt(DC, Left, Top, Right - Left, Bottom - Top, $00A000C9);
  SetBkColor(DC, SaveBkColor);
  SetTextColor(DC, SaveTextColor);
  DeleteObject(SelectObject(DC, SaveBrush));
  DeleteObject(Bitmap);
end;

procedure TThreadForm.Paint;
begin
  inherited Paint;
  if FShowGrayed then
    ShadeRect(Canvas.Handle, GetClientRect);
end;


{ TPageForm -------------------------------------------------------------------}

procedure TPageForm.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
{ If the WinControl parent is defined, then the Parent of the form becomes
  the Owner of the Control, the handle of the parent window is the handle of
  the control (usually a notebook or a palnel) }
  if WinCtrl <> NIL then
    with Params do
    begin
      Parent := TWinControl(WinCtrl.Owner);
      WndParent := WinCtrl.Handle;
      Style := WS_CHILD OR WS_CLIPSIBLINGS OR WS_CLIPCHILDREN;
      Align := alClient;
    end;
end;


constructor TPageForm.Create(AOwner: TComponent);
begin
{ if the owner is a TWinControl, get the Owner from it }
  if (AOwner is TWinControl) and not (AOwner is TForm) then
  begin
    WinCtrl := TWinControl(AOwner);
    inherited Create(WinCtrl.Owner);
  end
  else { nothing happens, it's a normal form }
  begin
    WinCtrl := NIL;
    inherited Create(AOwner);
  end;
end;

procedure TPageForm.OnUM_SETFOCUS(var message: TMessage);
var
  WinControl: TWinControl;
begin
  WinControl := FindNextControl(nil, True, True, False);
  if Assigned(WinControl) then
    if WinControl.CanFocus then
      WinControl.SetFocus;
end;


procedure TPageFormInfo.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
  begin
    if (AComponent = FFormContainer) then
      FFormContainer := nil;
  end;
end;

function TPageFormInfo.CreateForm: TPageForm;
begin
  if not Assigned(FormCreated) then
  begin
    if (FormType = nil) then
    begin
      result := nil;
      exit;
    end;
    FormCreated := FormType.Create(FormContainer);
    if Assigned(FormCreated) then
    begin
      FormDestroy := FormCreated.OnDestroy;
      FormCreated.OnDestroy := OnFormDestroy;
    end;
  end;
  result := FormCreated;
end;

procedure TPageFormInfo.HideForm;
begin
  if Assigned(FormCreated) then
    FormCreated.Hide;
end;

procedure TPageFormInfo.ShowForm;
begin
  if Assigned(FormCreated) then
  begin
    if FormCreated.Visible then
      FormCreated.Hide;
    FormCreated.Show;
{  Focus: Maus noch gedrckt, nichts funktioniert... }
    PostMessage(FormCreated.Handle, UM_SETFOCUS, 0, 0);

{    FormCreated.SetFocus;
    FormCreated.FindNextControl(nil, True, True, False).SetFocus;
    Application.ProcessMessages; }
{    FormCreated.ActiveControl := FormCreated.FindNextControl(nil, True, True, False); {}
  end;
end;

procedure TPageFormInfo.DestroyForm;
begin
  try
    if Assigned(FormCreated) then
      if not (csDestroying in FormCreated.ComponentState) then
        FormCreated.Free;
  finally
    FormCreated := nil;
  end;
end;

procedure TPageFormInfo.OnFormDestroy(Sender: TObject);
begin
  if Assigned(FormDestroy) then FormDestroy(Sender);
  FormDestroy := nil;
  FormCreated := nil;
end;




Constructor TFormPages.Create(AOwner:Tcomponent);
begin
  inherited Create(AOwner);
  FillChar(FormInfo, SizeOf(FormInfo), 0);
end;

destructor TFormPages.Destroy;
{var
  iPageNum: Integer; }
begin
(*
  try { Deleting all the forms created: }
    for iPageNum := 0 to MAXNBKPAGES - 1 do
      DestroyForm(iPageNum);
  finally
  end;
*)
  inherited Destroy;
end;

procedure TFormPages.Notification(AComponent: TComponent;
  Operation: TOperation);
var
  i: Integer;
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
  begin
    for i := 0 to MAXNBKPAGES - 1 do
      if (AComponent = FormInfo[i]) then
        FormInfo[i] := nil;
  end;
end;

procedure TFormPages.AddPage(Page: Integer; AFormContainer: TComponent;
  AFormType: TPageFormClass);
begin
  FormInfo[Page] := TPageFormInfo.Create(self);
  with FormInfo[Page] do
  begin
    FormType := AFormType;
    FormContainer := AFormContainer;
  end;
end;

function TFormPages.CreateForm(Page: Integer): TPageForm;
begin
  if (Page < 0) or (Page >= MAXNBKPAGES) or (not Assigned(FormInfo[Page])) then
    result := nil
  else
    result := FormInfo[Page].CreateForm;
end;

procedure TFormPages.HideForm(Page: Integer);
begin
  if (Page < 0) or (Page >= MAXNBKPAGES) or (not Assigned(FormInfo[Page])) then
    exit;
  FormInfo[Page].HideForm;
end;

procedure TFormPages.ShowForm(Page: Integer);
begin
  if (Page < 0) or (Page >= MAXNBKPAGES) or (not Assigned(FormInfo[Page])) then
    exit;
  FormInfo[Page].ShowForm;
end;

procedure TFormPages.DestroyForm(Page: Integer);
begin
  if (Page < 0) or (Page >= MAXNBKPAGES) or (not Assigned(FormInfo[Page])) then
    exit;
  FormInfo[Page].DestroyForm;
end;

procedure TFormPages.ChangeForm(Handle: HWND; OldPage, NewPage: Integer);
var
  OldCursor: HCursor;
begin
  if (OldPage = NewPage) then
    exit;
  OldCursor := SetCursor(LoadCursor(0, IDC_WAIT));
  LockWindowUpdate(Handle);
  if Assigned(FormInfo[OldPage]) then
    if FormInfo[OldPage].DontDestroy then
      HideForm(OldPage)
    else
      DestroyForm(OldPage);
  if (CreateForm(NewPage) <> nil) then
    ShowForm(NewPage);
  LockWindowUpdate(0);
  SetCursor(OldCursor);
end;


constructor TTabbedNotebookExternalPages.Create(AOwner:Tcomponent);
begin
  inherited Create(AOwner);
  FillChar(FormInfo, SizeOf(FormInfo), 0);
  ParentOnShow := nil;
  if Assigned(AOwner) and (AOwner is TForm) and
     not (csDesigning in ComponentState) then
  begin
    ParentOnShow := (AOwner as TForm).OnShow;
    (AOwner as TForm).OnShow := OnParentShow;
  { minor bug (glitch) - sets on pagecontrol too, if both notebook and
    pagecontrol used !! }
  end;
end;

procedure TTabbedNotebookExternalPages.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
  begin
    if (AComponent = FTabbedNotebook) then
      FTabbedNotebook := nil;
  end;
end;

function TTabbedNotebookExternalPages.GetTabbedNotebook: TTabbedNotebook;
begin
  Result := FTabbedNotebook;
end;

procedure TTabbedNotebookExternalPages.SetTabbedNotebook(value : TTabbedNotebook);
begin
  if Assigned(FTabbedNotebook) then
    ReleaseTabbedNotebook;
  FTabbedNotebook := Value;
  if Assigned(FTabbedNotebook) and
     not (csDesigning in ComponentState) then
  begin
    FTabbedNotebookChange := FTabbedNotebook.OnChange;
    FTabbedNotebook.OnChange := TabbedNotebookChange;
  end
  else
    FTabbedNotebookChange := nil;
end;

procedure TTabbedNotebookExternalPages.ReleaseTabbedNotebook;
begin
  if Assigned(FTabbedNotebook) then
  begin
{   if not (csDestroying in FTabbedNotebook.ComponentState) then }
    FTabbedNotebook.OnChange := FTabbedNotebookChange;
  end;
  FTabbedNotebook := nil;
  FTabbedNotebookChange := nil;
end;

procedure TTabbedNotebookExternalPages.TabbedNotebookChange(Sender: TObject;
  NewTab: Integer; var AllowChange: Boolean);
begin
  if Assigned(FTabbedNotebookChange) then
    FTabbedNotebookChange(Sender, NewTab, AllowChange);
  if AllowChange and FTabbedNotebook.Visible then
  begin
    ChangeForm(FTabbedNotebook.Handle, FTabbedNotebook.PageIndex, NewTab);
{
    if Assigned(Owner) and (Owner is TForm) and
       not (csDesigning in ComponentState) then
      TForm(Owner).ActiveControl := CreateForm(NewTab); {}
  end;
end;

procedure TTabbedNotebookExternalPages.CreateCurrentPage;
begin
  if not Assigned(FTabbedNotebook) then
    exit;
  if (CreateForm(FTabbedNotebook.PageIndex) <> nil) then
    ShowForm(FTabbedNotebook.PageIndex);
end;

procedure TTabbedNotebookExternalPages.OnParentShow(Sender: TObject);
begin
  if Assigned(ParentOnShow) then
    ParentOnShow(Sender);
  CreateCurrentPage; {}
end;



{ TPageControlExternalPages ------------------------------------------------ }

constructor TPageControlExternalPages.Create(AOwner:Tcomponent);
begin
  inherited Create(AOwner);
  FillChar(FormInfo, SizeOf(FormInfo), 0);
  ParentOnShow := nil;
  if Assigned(AOwner) and (AOwner is TForm) and
     not (csDesigning in ComponentState) then
  begin
    ParentOnShow := (AOwner as TForm).OnShow;
    (AOwner as TForm).OnShow := OnParentShow;
  end;
end;

procedure TPageControlExternalPages.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
  begin
    if (AComponent = FPageControl ) then
    begin
      FPageControl  := nil;       // called if exiting
   end;
  end;
end;

function TPageControlExternalPages.GeTPageControl : TPageControl ;
begin
  Result := FPageControl ;
end;

procedure TPageControlExternalPages.SeTPageControl (value : TPageControl );
begin
  if Assigned(FPageControl ) then
    ReleasePageControl ;
  FPageControl  := Value;
  if Assigned(FPageControl ) and
  not (csDesigning in ComponentState) then
    begin
    FPageControlChange := FPageControl.OnChange;
    FPageControl.OnChange := PageControlChange;

    FPageControlChanging := PageControl.OnChanging ;
    FPageControl.OnChanging := PageControlChanging ;

  end
  else
    FPageControlChange := nil;
end;

procedure TPageControlExternalPages.ReleasePageControl ;
begin
  if Assigned(FPageControl ) then
  begin
{   if not (csDestroying in FPageControl .ComponentState) then }  // inherited bug !! //
  FPageControlChange := nil;
  FPageControlChanging := nil;
  FPageControl  := nil;

  end;
end;

procedure TPageControlExternalPages.PageControlChanging(Sender: TObject ;
   var AllowChange: Boolean);   // NewTab: Integer;);   // NewTab: Integer;  var AllowChange: Boolean
begin
if  not (csDesigning in ComponentState) then
begin
  if Assigned(FPageControlChanging) then
    FPageControlChanging(Sender, AllowChange);

  if AllowChange and FPageControl.Visible then
     PreviousPage :=  FPageControl.ActivePage ; // set old page on changing !!

end;
end;


procedure TPageControlExternalPages.PageControlChange(Sender: TObject) ;

begin
if  not (csDesigning in ComponentState) then
begin
  if Assigned(FPageControlChange) then
    FPageControlChange(Sender);

  if (PreviousPage <> nil)        // set onchanging !!
  and (PreviousPage <>  FPageControl.ActivePage) then   // move to notify or ???
     begin
     ChangeForm(FPageControl.Handle,  PreviousPage.PageIndex , FPageControl.ActivePage.PageIndex );
     end ;
  end;
end;

procedure TPageControlExternalPages.CreateCurrentPage;
begin
  if not Assigned(FPageControl ) then
    exit;
  if (CreateForm(FPageControl.ActivePage.PageIndex) <> nil) then
    ShowForm(FPageControl.ActivePage.PageIndex);
end;

procedure TPageControlExternalPages.OnParentShow(Sender: TObject);
begin
  if Assigned(ParentOnShow) then
    ParentOnShow(Sender);
  CreateCurrentPage; {}
end;


{ TPseudoThread ----------------------------------------------------------- }

constructor TPseudoThread.Create(CreateSuspended: Boolean);
begin
  inherited Create;
  FTerminated := false;
  FSuspended := CreateSuspended;
  if not FSuspended then
    Execute;
end;

procedure TPseudoThread.Synchronize(Method: TThreadMethod);
begin
  Method;
end;

procedure TPseudoThread.Resume;
begin
  if FSuspended then
  begin
    Execute;
    FSuspended := false;
  end;
end;

procedure TPseudoThread.Suspend;
begin
end;

procedure TPseudoThread.Terminate;
begin
  if FFreeOnTerminate and not FTerminated then
  begin
    FTerminated := true;
    Destroy;
  end
  else
    FTerminated := true;
end;




procedure Register;
begin
  RegisterComponents('Samples',
    [TPageForm, TPageFormInfo, TTabbedNotebookExternalPages,
     TPageControlExternalPages]);
end;

end.
